@@ -12,7 +12,7 @@ use Log::Log4perl;
 use Module::Load;
 use Crypt::LE ':errors', ':keys';
 
-my $VERSION = '0.19';
+my $VERSION = '0.20';
 
 use constant PEER_CRT  => 4;
 use constant CRT_DEPTH => 5;
@@ -24,8 +24,8 @@ sub main {
     my $opt = { logger => Log::Log4perl->get_logger() };
     binmode(STDOUT, ":encoding(UTF-8)");
     if (my $rv = work($opt)) {
-        $opt->{logger}->error($rv);
-        return 255;
+        $opt->{logger}->error($rv->{'msg'}) if $rv->{'msg'};
+        return defined $rv->{'code'} ? $rv->{'code'} : 255;
     }
     return 0;
 }
@@ -39,18 +39,18 @@ sub work {
 
     if (-r $opt->{'key'}) {
         $opt->{'logger'}->info("Loading an account key from $opt->{'key'}");
-        $le->load_account_key($opt->{'key'}) == OK or return "Could not load an account key: " . $le->error_details;
+        $le->load_account_key($opt->{'key'}) == OK or return _error("Could not load an account key: " . $le->error_details);
     } else {
         $opt->{'logger'}->info("Generating a new account key");
-        $le->generate_account_key == OK or return "Could not generate an account key: " . $le->error_details;
+        $le->generate_account_key == OK or return _error("Could not generate an account key: " . $le->error_details);
         $opt->{'logger'}->info("Saving generated account key into $opt->{'key'}");
-        return "Failed to save an account key file" if _write($opt->{'key'}, $le->account_key);
+        return _error("Failed to save an account key file") if _write($opt->{'key'}, $le->account_key);
     }
 
     if ($opt->{'revoke'}) {
-        return "Name of the certificate file should be specified." unless $opt->{'crt'};
+        return _error("Name of the certificate file should be specified.") unless $opt->{'crt'};
         my $crt = _read($opt->{'crt'});
-        return "Could not read the certificate file." unless $crt;
+        return _error("Could not read the certificate file.") unless $crt;
         # Take the first certificate in file, disregard the issuer's one.
         $crt=~s/^(.*?-+\s*END CERTIFICATE\s*-+).*/$1/s;
         my $rv = $le->revoke_certificate(\$crt);
@@ -59,29 +59,30 @@ sub work {
         } elsif ($rv == ALREADY_DONE) {
             $opt->{'logger'}->info("Certificate has been ALREADY revoked.");
         } else {
-            return "Problem with revoking certificate: " . $le->error_details;
+            return _error("Problem with revoking certificate: " . $le->error_details);
         }
         return;
     }
 
     if (-r $opt->{'csr'}) {
         $opt->{'logger'}->info("Loading a CSR from $opt->{'csr'}");
-        $le->load_csr($opt->{'csr'}, $opt->{'domains'}) == OK or return "Could not load a CSR: " . $le->error_details;
+        $le->load_csr($opt->{'csr'}, $opt->{'domains'}) == OK or return _error("Could not load a CSR: " . $le->error_details);
     } else {
         $opt->{'logger'}->info("Generating a new CSR for domains $opt->{'domains'}");
         if (-e $opt->{'csr-key'}) {
              # Allow using pre-existing key when generating CSR
-             return "Could not load existing CSR key from $opt->{'csr-key'}" if $le->load_csr_key($opt->{'csr-key'});
+             return _error("Could not load existing CSR key from $opt->{'csr-key'} - " . $le->error_details) if $le->load_csr_key($opt->{'csr-key'});
              $opt->{'logger'}->info("New CSR will be based on '$opt->{'csr-key'}' key");
         } else {
              $opt->{'logger'}->info("New CSR will be based on a generated key");
         }
-        $le->generate_csr($opt->{'domains'}, KEY_RSA, $opt->{'legacy'} ? 2048 : 4096) == OK or return "Could not generate a CSR: " . $le->error_details;
+        my ($type, $attr) = $opt->{'curve'} ? (KEY_ECC, $opt->{'curve'}) : (KEY_RSA, $opt->{'legacy'} ? 2048 : 4096);
+        $le->generate_csr($opt->{'domains'}, $type, $attr) == OK or return _error("Could not generate a CSR: " . $le->error_details);
         $opt->{'logger'}->info("Saving a new CSR into $opt->{'csr'}");
         return "Failed to save a CSR" if _write($opt->{'csr'}, $le->csr);
         unless (-e $opt->{'csr-key'}) {
             $opt->{'logger'}->info("Saving a new CSR key into $opt->{'csr-key'}");
-            return "Failed to save a CSR key" if _write($opt->{'csr-key'}, $le->csr_key);
+            return _error("Failed to save a CSR key") if _write($opt->{'csr-key'}, $le->csr_key);
         }
     }
 
@@ -106,7 +107,7 @@ sub work {
                 }
             }
         }
-        return "Could not get the certificate expiration value, cannot renew." if $rv;
+        return _error("Could not get the certificate expiration value, cannot renew.") if $rv;
         if ($opt->{'expires'} > $opt->{'renew'}) {
             $opt->{'logger'}->info("Too early for renewal, certificate expires in $opt->{'expires'} days.");
             return;
@@ -115,11 +116,11 @@ sub work {
     }
     
     if ($opt->{'email'}) {
-        return $le->error_details if $le->set_account_email($opt->{'email'});
+        return _error($le->error_details) if $le->set_account_email($opt->{'email'});
     }
     
     $opt->{'logger'}->info("Registering the account key");
-    return $le->error_details if $le->register;
+    return _error($le->error_details) if $le->register;
     my $current_account_id = $le->registration_id||'unknown';
     $opt->{'logger'}->info($le->new_registration ? "The key has been successfully registered. ID: $current_account_id" : "The key is already registered. ID: $current_account_id");
     $opt->{'logger'}->info("Make sure to check TOS at " . $le->tos) if ($le->tos_changed and $le->tos);
@@ -130,27 +131,27 @@ sub work {
         $opt->{'logger'}->info("Received domain certificate, no validation required at this time.");
     } else {
         # If it's not an auth problem, but blacklisted domains for example - stop.
-        return "Error requesting certificate: " . $le->error_details if $new_crt_status != AUTH_ERROR;
-        return $le->error_details if $le->request_challenge();
-        return $le->error_details if $le->accept_challenge($opt->{'handler'} || \&process_challenge, $opt->{'handle-params'}, $opt->{'handle-as'});
-        return $le->error_details if $le->verify_challenge($opt->{'handler'} || \&process_verification, $opt->{'handle-params'}, $opt->{'handle-as'});
+        return _error("Error requesting certificate: " . $le->error_details) if $new_crt_status != AUTH_ERROR;
+        return _error($le->error_details) if $le->request_challenge();
+        return _error($le->error_details) if $le->accept_challenge($opt->{'handler'} || \&process_challenge, $opt->{'handle-params'}, $opt->{'handle-as'});
+        return _error($le->error_details) if $le->verify_challenge($opt->{'handler'} || \&process_verification, $opt->{'handle-params'}, $opt->{'handle-as'});
     }
     unless ($le->certificate) {
         $opt->{'logger'}->info("Requesting domain certificate.");
-        return $le->error_details if $le->request_certificate();
+        return _error($le->error_details) if $le->request_certificate();
     }
     $opt->{'logger'}->info("Requesting issuer's certificate.");
     if ($le->request_issuer_certificate()) {
         $opt->{'logger'}->error("Could not download an issuer's certificate, try to download manually from " . $le->issuer_url);
         $opt->{'logger'}->warn("Will be saving the domain certificate alone, not the full chain.");
-        return "Failed to save the domain certificate file" if _write($opt->{'crt'}, $le->certificate);
+        return _error("Failed to save the domain certificate file") if _write($opt->{'crt'}, $le->certificate);
     } else {
         unless ($opt->{'legacy'}) {
             $opt->{'logger'}->info("Saving the full certificate chain to $opt->{'crt'}.");
-            return "Failed to save the domain certificate file" if _write($opt->{'crt'}, $le->certificate . "\n" . $le->issuer . "\n");
+            return _error("Failed to save the domain certificate file") if _write($opt->{'crt'}, $le->certificate . "\n" . $le->issuer . "\n");
         } else {
             $opt->{'logger'}->info("Saving the domain certificate to $opt->{'crt'}.");
-            return "Failed to save the domain certificate file" if _write($opt->{'crt'}, $le->certificate);
+            return _error("Failed to save the domain certificate file") if _write($opt->{'crt'}, $le->certificate);
             $opt->{'crt'}=~s/\.[^\.]+$//;
             $opt->{'crt'}.='.ca';
             $opt->{'logger'}->info("Saving the issuer's certificate to $opt->{'crt'}.");
@@ -168,21 +169,22 @@ sub work {
         $rv = $opt->{'complete-handler'}->complete($data, $opt->{'complete-params'});
         };
         if ($@ or !$rv) {
-            return "Completion handler " . ($@ ? "thrown an error: $@" : "did not return a true value");
+            return _error("Completion handler " . ($@ ? "thrown an error: $@" : "did not return a true value"));
         }
     }
 
+    $opt->{'logger'}->info("===> NOTE: You have been using the test server for this certificate. To issue a valid trusted certificate add --live option.") unless $opt->{'live'};
     $opt->{'logger'}->info("The job is done, enjoy your certificate! For feedback and bug reports contact us at [ https://ZeroSSL.com | https://Do-Know.com ]\n");
-    return;
+    return { code => $opt->{'issue-code'}||0 };
 }
 
 sub parse_options {
     my $opt = shift;
     my $args = @ARGV;
     
-    GetOptions ($opt, 'key=s', 'csr=s', 'csr-key=s', 'domains=s', 'path=s', 'crt=s', 'email=s', 'renew=i',
-            'handle-with=s', 'handle-as=s', 'handle-params=s', 'complete-with=s', 'complete-params=s', 'log-config=s',
-            'generate-missing', 'generate-only', 'revoke', 'legacy', 'unlink', 'live', 'debug', 'help') || return "Use --help to see the usage examples.\n";
+    GetOptions ($opt, 'key=s', 'csr=s', 'csr-key=s', 'domains=s', 'path=s', 'crt=s', 'email=s', 'curve=s', 'renew=i',
+            'issue-code=i', 'handle-with=s', 'handle-as=s', 'handle-params=s', 'complete-with=s', 'complete-params=s', 'log-config=s',
+            'generate-missing', 'generate-only', 'revoke', 'legacy', 'unlink', 'live', 'debug', 'help') || return _error("Use --help to see the usage examples.");
             
     usage_and_exit() unless ($args and !$opt->{'help'});
     my $rv = reconfigure_log($opt);
@@ -190,34 +192,34 @@ sub parse_options {
 
     $opt->{'logger'}->info("[ ZeroSSL Crypt::LE client v$VERSION started. ]");
 
-    return "Incorrect parameters - need account key file name specified." unless $opt->{'key'};
+    return _error("Incorrect parameters - need account key file name specified.") unless $opt->{'key'};
     if (-e $opt->{'key'}) {
-        return "Account key file is not readable." unless (-r $opt->{'key'});
+        return _error("Account key file is not readable.") unless (-r $opt->{'key'});
     } else {
-        return "Account key file is missing and the option to generate missing files is not used." unless $opt->{'generate-missing'};
+        return _error("Account key file is missing and the option to generate missing files is not used.") unless $opt->{'generate-missing'};
     }
 
     unless ($opt->{'crt'} or $opt->{'generate-only'}) {
-        return "Please specify a file name for the certificate.";
+        return _error("Please specify a file name for the certificate.");
     }
 
     if ($opt->{'revoke'}) {
-        return "Need a certificate file for revoke to work." unless (-r $opt->{'crt'});
-        return "Need an account key - revoke assumes you had a registered account when got the certificate." unless (-r $opt->{'key'});
+        return _error("Need a certificate file for revoke to work.") unless (-r $opt->{'crt'});
+        return _error("Need an account key - revoke assumes you had a registered account when got the certificate.") unless (-r $opt->{'key'});
     } else {
-        return "Incorrect parameters - need CSR file name specified." unless $opt->{'csr'};
+        return _error("Incorrect parameters - need CSR file name specified.") unless $opt->{'csr'};
         if (-e $opt->{'csr'}) {
-            return "CSR file is not readable." unless (-r $opt->{'csr'});
+            return _error("CSR file is not readable.") unless (-r $opt->{'csr'});
         } else {
-            return "CSR file is missing and the option to generate missing files is not used." unless $opt->{'generate-missing'};
-            return "CSR file is missing and CSR-key file name is not specified." unless $opt->{'csr-key'};
-            return "Domain list should be provided to generate a CSR." unless ($opt->{'domains'} and $opt->{'domains'}!~/^[\s\,]*$/);
+            return _error("CSR file is missing and the option to generate missing files is not used.") unless $opt->{'generate-missing'};
+            return _error("CSR file is missing and CSR-key file name is not specified.") unless $opt->{'csr-key'};
+            return _error("Domain list should be provided to generate a CSR.") unless ($opt->{'domains'} and $opt->{'domains'}!~/^[\s\,]*$/);
         }
 
         if ($opt->{'path'}) {
-            return "Path to save challenge files into should be a writable directory" unless (-d $opt->{'path'} and -w _);
+            return _error("Path to save challenge files into should be a writable directory.") unless (-d $opt->{'path'} and -w _);
         } elsif ($opt->{'unlink'}) {
-            return "Unlink option will have no effect without --path.";
+            return _error("Unlink option will have no effect without --path.");
         }
 
         $opt->{'handle-as'} = $opt->{'handle-as'} ? lc($opt->{'handle-as'}) : 'http';
@@ -227,9 +229,9 @@ sub parse_options {
                 load $opt->{'handle-with'};
                 $opt->{'handler'} = $opt->{'handle-with'}->new();
             };
-            return "Cannot use the module to handle challenges with." if $@;
+            return _error("Cannot use the module to handle challenges with.") if $@;
             my $method = 'handle_challenge_' . $opt->{'handle-as'};
-            return "Module to handle challenges does not seem to support the challenge type of $opt->{'handle-as'}." unless $opt->{'handler'}->can($method);
+            return _error("Module to handle challenges does not seem to support the challenge type of $opt->{'handle-as'}.") unless $opt->{'handler'}->can($method);
             my $rv = _load_params($opt, 'handle-params');
             return $rv if $rv;
         } else {
@@ -241,8 +243,8 @@ sub parse_options {
                 load $opt->{'complete-with'};
                 $opt->{'complete-handler'} = $opt->{'complete-with'}->new();
             };
-            return "Cannot use the module to complete processing with." if $@;
-            return "Module to complete processing with does not seem to support the required 'complete' method." unless $opt->{'complete-handler'}->can('complete');
+            return _error("Cannot use the module to complete processing with.") if $@;
+            return _error("Module to complete processing with does not seem to support the required 'complete' method.") unless $opt->{'complete-handler'}->can('complete');
             my $rv = _load_params($opt, 'complete-params');
             return $rv if $rv;
         } else {
@@ -266,7 +268,7 @@ sub reconfigure_log {
         };
         if ($@ or !%{Log::Log4perl::appenders()}) {
             Log::Log4perl->easy_init();
-            return "Could not init logging with '$opt->{'log-config'}' file";
+            return _error("Could not init logging with '$opt->{'log-config'}' file");
         }
         $opt->{logger} = Log::Log4perl->get_logger();
     }
@@ -278,14 +280,14 @@ sub _load_params {
     return unless ($opt and $opt->{$type});
     if ($opt->{$type}!~/[\{\[\}\]]/) {
         $opt->{$type} = _read($opt->{$type});
-        return "Could not read the file with '$type'." unless $opt->{$type};
+        return _error("Could not read the file with '$type'.") unless $opt->{$type};
     }
     my $j = JSON->new->canonical()->allow_nonref();
     eval {
         $opt->{$type} = $j->decode($opt->{$type});
     };
     return ($@ or (ref $opt->{$type} ne 'HASH')) ? 
-        "Could not decode '$type'. Please make sure you are providing a valid JSON document and {} are in place." . ($opt->{'debug'} ? $@ : '') : 0;
+        _error("Could not decode '$type'. Please make sure you are providing a valid JSON document and {} are in place." . ($opt->{'debug'} ? $@ : '')) : 0;
 }
 
 sub _read {
@@ -310,6 +312,11 @@ sub _write {
     return 0;
 }
 
+sub _error {
+    my ($msg, $code) = @_;
+    return { msg => $msg, code => $code||255 };
+}
+
 sub verify_crt {
     my $opt = shift;
     return sub {
@@ -384,7 +391,7 @@ sub process_verification {
 
 __END__
 
- ZeroSSL Crypt::LE client v0.19
+ ZeroSSL Crypt::LE client v0.20
 
  ===============
  USAGE EXAMPLES: 
@@ -473,6 +480,7 @@ methods:
  key <file>                       - Your account key file.
  csr <file>                       - Your CSR file.
  csr-key <file>                   - Key for your CSR (only mandatory if CSR is missing and to be generated).
+ curve <name|default>             - Use ECC with the specified curve instead of RSA to generate a domain key and CSR.
  domains <list>                   - Domains as comma-separated list (only mandatory if CSR is missing).
  path <absolute path>             - Path to local .well-known/acme-challenge/ to drop required challenge files into (optional).
  handle-with <Some::Module>       - Module name to handle challenges with (optional).
@@ -480,6 +488,7 @@ methods:
  handle-params <{json}|file>      - JSON (or name of the file containing it) with parameters to be passed to the challenge-handling module (optional).
  complete-with <Another::Module>  - Module name to handle process completion with (optional).
  complete-params <{json}|file>    - JSON (or name of the file containing it) with parameters to be passed to the completion-handling module (optional).
+ issue-code XXX                   - Use specific exit code on certificate issuance/renewal.
  email <some@mail.address>        - Mail address for the account registration and certificate expiration notifications (optional).
  log-config <file>                - Configuration file for logging (perldoc Log::Log4perl to see configuration examples).
  generate-missing                 - Generate missing files (key, csr and csr-key).
